package util;

import java.util.List;

import junit.framework.TestCase;

import org.eclipse.emf.codegen.util.ImportManager;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EParameter;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.ETypedElement;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;

public class EcoreGenerationHelperTest extends TestCase {
    
    private EcoreGenerationHelper helper;
	private ImportManager importManager;
    
	private EClass companyClass;
	private EClass employeeClass;
	private EClass departmentClass;
	private EPackage companyPackage;
	private EOperation salaryOperation;
    
    private EClass base;
    private EClass interface1;
    private EClass interface2;
    private EClass interface3;
    
    private EPackage otherPackage;
    private EClass otherClass;

	protected void setUp() throws Exception {
        helper = new EcoreGenerationHelper();
		importManager = helper.makeImportManager("");
		setupCompany();
	}
	
	private void setupCompany() {
		EcoreFactory ecoreFactory = EcoreFactory.eINSTANCE;
		EcorePackage ecorePackage = EcorePackage.eINSTANCE;

//		 create an Company class
		companyClass = ecoreFactory.createEClass();
		companyClass.setName("Company");
        
        EAnnotation companyAnnotation = ecoreFactory.createEAnnotation();
        companyAnnotation.getDetails().put("a", "b");
        companyAnnotation.setEModelElement(companyClass);
        
        base = ecoreFactory.createEClass();
        base.setName("base");
        companyClass.getESuperTypes().add(base);
        interface1 = ecoreFactory.createEClass();
        interface1.setName("interface1");
        interface1.setInterface(true);
        companyClass.getESuperTypes().add(interface1);
        interface2 = ecoreFactory.createEClass();
        interface2.setName("interface2");
        interface2.setInterface(true);
        companyClass.getESuperTypes().add(interface2);
        interface3 = ecoreFactory.createEClass();
        interface3.setName("interface3");
        interface3.setInterface(true);
        interface2.getESuperTypes().add(interface3);
        
        
		
//		 create company name
		EAttribute companyName = ecoreFactory.createEAttribute();
		companyName.setName("name");
		companyName.setEType(ecorePackage.getEString());
		companyClass.getEStructuralFeatures().add(companyName);
		
		salaryOperation = ecoreFactory.createEOperation();
		salaryOperation.setName("paySalary");
		salaryOperation.setEType(ecorePackage.getEDate());
		EParameter salaryParam1 = ecoreFactory.createEParameter();
		salaryParam1.setName("salaryBonus");
		salaryParam1.setEType(ecorePackage.getEBigDecimal());
		salaryOperation.getEParameters().add(salaryParam1);
        EParameter salaryParam2 = ecoreFactory.createEParameter();
        salaryParam2.setName("helloText");
        salaryParam2.setEType(ecorePackage.getEString());
        salaryOperation.getEParameters().add(salaryParam2);
        companyClass.getEOperations().add(salaryOperation);

//		create an Employee class
		employeeClass = ecoreFactory.createEClass();
		employeeClass.setName("Employee");

//		add a name attribute to an Employee class
		EAttribute employeeName = ecoreFactory.createEAttribute();
		employeeName.setName("name");
		employeeName.setEType(ecorePackage.getEString());
		employeeClass.getEStructuralFeatures().add(employeeName);

//		create a Department class
		departmentClass = ecoreFactory.createEClass();
		departmentClass.setName("Department");

//		add department identification number
		EAttribute departmentNumber = ecoreFactory.createEAttribute();
		departmentNumber.setName("number");
		departmentNumber.setEType(ecorePackage.getEInt());
		departmentClass.getEStructuralFeatures().add(departmentNumber);

//		department class can contain reference to one or many employees
		EReference departmentEmployees = ecoreFactory.createEReference();
		departmentEmployees.setName("employees");
		departmentEmployees.setEType(employeeClass);
//		 specify that it could be one or more employees
		departmentEmployees.setUpperBound(ETypedElement.UNBOUNDED_MULTIPLICITY);
		departmentEmployees.setContainment(true);
		departmentClass.getEStructuralFeatures().add(departmentEmployees);
        EReference employeeDepartment = ecoreFactory.createEReference();
        employeeDepartment.setUpperBound(1);
        employeeDepartment.setName("department");
        employeeDepartment.setEType(departmentClass);
        employeeClass.getEStructuralFeatures().add(employeeDepartment);

//		 company can contain reference to one or more departments
		EReference companyDepartments = ecoreFactory.createEReference();
		companyDepartments.setName("department");
		companyDepartments.setEType(departmentClass);
		companyDepartments.setUpperBound(ETypedElement.UNBOUNDED_MULTIPLICITY);
		companyDepartments.setContainment(true);
		companyClass.getEStructuralFeatures().add(companyDepartments);

//		create a package that represents company
        EPackage com = ecoreFactory.createEPackage();
        com.setName("com");
		companyPackage = ecoreFactory.createEPackage();
		companyPackage.setName("foocompany");
		companyPackage.setNsPrefix("foo");
		companyPackage.setNsURI("http:///com.example.company.ecore");
        com.getESubpackages().add(companyPackage);
		companyPackage.getEClassifiers().add(employeeClass);
		companyPackage.getEClassifiers().add(departmentClass);
		companyPackage.getEClassifiers().add(companyClass);
        
        otherClass = ecoreFactory.createEClass();
        otherClass.setName("Other");
        otherPackage = ecoreFactory.createEPackage();
        otherPackage.setName("other");
        otherPackage.getEClassifiers().add(otherClass);
	}

	public void testGetNameClass() {
		String name = helper.getName(companyClass);
		assertEquals("Company", name);
    }
    
    public void testGetQualifiedNameClass() {
        String name = helper.getQualifiedName(companyClass);
        assertEquals("com.foocompany.Company", name);
    }
    
    public void testGetNamePackage() {
        String name = helper.getName(companyPackage);
        assertEquals("foocompany", name);
	}
    
    public void testGetQualifiedNamePackage() {
        String name = helper.getQualifiedName(companyPackage);
        assertEquals("com.foocompany", name);
    }
    
    public void testGetTypeNameOperation() {
        String name = helper.getTypeName(salaryOperation);
        // TODO borde vara Date (java.util.Date)
        assertEquals("Date", name);
    }
    
    public void testGetTypeQualifiedNameOperation() {
        String name = helper.getTypeQualifiedName(salaryOperation);
        // TODO borde vara java.util.Date
        assertEquals("java.util.Date", name);
    }
    
    public void testGetTypeNameParameter() {
        String name = helper.getTypeName((EParameter) salaryOperation.getEParameters().get(0));
        // TODO borde vara BigDecimal (java.math.BigDecimal)
        assertEquals("BigDecimal", name);
    }
    
    public void testGetTypeQualifiedNameParameter() {
        String name = helper.getTypeQualifiedName((EParameter) salaryOperation.getEParameters().get(0));
        // TODO borde vara java.math.BigDecimal
        assertEquals("java.math.BigDecimal", name);
    }
    
    public void testGetTypeNameAttribute() {
        String name = helper.getTypeName((EAttribute) departmentClass.getEAttributes().get(0));
        // TODO borde vara int 
        assertEquals("int", name);
    }
    
    public void testGetTypeQualifiedNameAttribute() {
        String name = helper.getTypeQualifiedName((EAttribute) departmentClass.getEAttributes().get(0));
        // TODO borde vara int
        assertEquals("int", name);
    }
    
    public void testExtends() {
        List<EClass> list = helper.getExtends(companyClass);
        assertEquals(1, list.size());
        assertSame(base, list.get(0));
        
        list = helper.getExtends(interface2);
        assertEquals(1, list.size());
        assertSame(interface3, list.get(0));
    }
    
    public void testImplements() {
        List<EClass> list = helper.getImplements(companyClass);
        assertEquals(2, list.size());
        assertSame(interface1, list.get(0));
        assertSame(interface2, list.get(1));
        
        list = helper.getImplements(interface2);
        assertEquals(1, list.size());
        assertSame(interface3, list.get(0));
    }
    
    public void testGetExtendsAndImplementsLitteral() {
        String s = helper.getExtendsAndImplementsLitteral(companyClass);
        assertEquals("extends base implements interface1, interface2", s);
    }
    
    public void testAllManyReferences() {
        List<EReference> list = helper.getAllManyReferences(departmentClass);
        assertEquals(1, list.size());
        assertEquals("employees", list.get(0).getName());
        assertSame(employeeClass, list.get(0).getEReferenceType());
        
        list = helper.getAllManyReferences(employeeClass);
        assertEquals(0, list.size());
    }
    
    public void testAllOneReferences() {
        List<EReference> list = helper.getAllOneReferences(employeeClass);
        assertEquals(1, list.size());
        assertEquals("department", list.get(0).getName());
        assertSame(departmentClass, list.get(0).getEReferenceType());
        
        list = helper.getAllOneReferences(departmentClass);
        assertEquals(0, list.size());
    }
    
    public void testGetAnnotation() {
        String value = helper.getAnnotation(companyClass, "a");
        assertEquals("b", value);
        
        value = helper.getAnnotation(companyClass, "none");
        assertNull(value);
    }
    
    public void testGetParameterList() {
        String s = helper.getParameterList(salaryOperation);
        assertEquals("BigDecimal salaryBonus, String helloText", s);
    }
    
    public void testImport() {
        helper.makeImportManager(companyPackage);
        
        helper.getImportManager().addImport("foo.Bar");
        helper.addImport("java.util.Map");
        
        helper.getName(companyClass);
        helper.getName(otherClass);
        
        String importString = helper.getImportManager().computeSortedImports();
        
        assertTrue(importString.indexOf("import foo.Bar;") != -1);
        assertTrue(importString.indexOf("import com.company.Company;") == -1);
        assertTrue(importString.indexOf("import other.Other;") != -1);
        assertTrue(importString.indexOf("import java.util.Map;") != -1);
        
    }

}
